from setuptools import setup, find_packages
setup(
name='rHDPE_Data_Analysis',
version='1.0.110',
author='Phil Smith',
author_email='pvsmith100@btinternet.com',
description='Package for analysing data relating to recycled HDPE',
packages=find_packages(),
classifiers=[
'Programming Language :: Python :: 3',
'License :: OSI Approved :: MIT License',
'Operating System :: OS Independent',
],
python_requires='>=3.6',
)
